--[[
Name: LibMath
Revision: $Rev: 0 $
Author(s): Humbedooh
Description: Extended math library for Lua
Optional dependencies: LibStub
License: MIT
]]
-----------------------------------------------------------------------

-- [[ UPVALUES ]]--
local _G = _G
local assert = _G.assert
local upper = assert(_G.string.upper)
local lower = assert(_G.string.lower)
local tonumber = assert(_G.tonumber)
local type = assert(_G.type)
local error = assert(_G.error)
local select = assert(_G.select)
local tsort = assert(_G.table.sort)
local floor = assert(_G.math.floor)
local ceil = assert(_G.math.ceil)
local log = assert(_G.math.log)
local sqrt = assert(_G.math.sqrt)
local pow = assert(_G.math.pow)
local tan = assert(_G.math.tan)
local cos = assert(_G.math.cos)
local sin = assert(_G.math.sin)
local math = assert(_G.math)

--[[ MATH SET ]]--

--- Returns the sum of all numbers passed
-- @param ... a list of numbers to be added up
-- @return the sum of the numbers.
function math.sum(...)
    local val = 0
    for k, v in pairs({...}) do
        val = val + (tonumber(v) or 0);
    end
    return val;
end

--- Returns the average (mean value) of the numbers passed.
function math.avg(...)
    local val = 0;
    local num = 0;
    for k, v in pairs({...}) do
        val = val + (tonumber(v) or 0);
        if ( tonumber(v) ~= nil ) then
            num = num + 1;
        end
    end
    return val/num;
end

--- Returns the most frequently occurring number among those passed.
function math.mode(...)
    local vals = {...};
    local mode = {};
    table.sort(vals);
    for k, v in pairs(vals) do
        mode[v] = (mode[v] or 0) + 1;
    end
    local n = 0;
    local m = 0;
    for k, v in pairs(mode) do
        if ( v > n ) then n = v; m = k; end
    end
    return m;
end

--- Returns the median (the number in the middle of all the numbers sorted numerically)
function math.median(...)
    local vals = {...};
    sort(vals);
    return vals[math.floor((#vals/2) + 0.5)] or 0;
end

--- Returns the k-th largest value among the numbers passed
function math.large(k, ...)
    local vals = {...};
    sort(vals);
    return vals[#vals-(k-1)] or 0;
end
--- Returns the k-th smallest value among the numbers passed
function math.small(k, ...)
    local vals = {...};
   sort(vals);
    return vals[k] or 0;
end

--- Returns the sum of all numbers that passes the statement (fx. ">25")
function math.sumif(statement, ...)
    local vals = {...};
	local ret = 0;
	if ( type(statement) == "string" ) then
		local i = 1;
		if ( statement:sub(2,2) == "=" ) then i = 2; end
		local sign, number = statement:sub(1,i), tonumber(statement:sub(i+1));
		for k, v in pairs(vals) do
			if ( sign == '=' and v == number ) then ret = ret + v; end
			if ( sign == '>' and v > number ) then ret = ret + v; end
			if ( sign == '<' and v < number ) then ret = ret + v; end
			if ( sign == '>=' and v >= number ) then ret = ret + v; end
			if ( sign == '<=' and v <= number ) then ret = ret + v; end
			if ( sign == '!=' and v ~= number ) then ret = ret + v; end
		end
	end
	return ret;
end

--- Returns the total number of numbers passed that fits the statement (fx. ">= 4")
function math.countif(statement, ...)
    local vals = {...};
	local ret = 0;
	if ( type(statement) == "string" ) then
		local i = 1;
		if ( statement:sub(2,2) == "=" ) then i = 2; end
		local sign, number = statement:sub(1,i), tonumber(statement:sub(i+1));
		for k, v in pairs(vals) do
			if ( sign == '=' and v == number ) then ret = ret + 1; end
			if ( sign == '>' and v > number ) then ret = ret + 1; end
			if ( sign == '<' and v < number ) then ret = ret + 1; end
			if ( sign == '>=' and v >= number ) then ret = ret + 1; end
			if ( sign == '<=' and v <= number ) then ret = ret + 1; end
			if ( sign == '!=' and v ~= number ) then ret = ret + 1; end
		end
	end
	return ret;
end

--- Returns -1 for numbers less than zero, +1 for numbers greater than zero and 0 for zero.
function math.sign(val)
    return ( val > 0 and 1 ) or ( val < 0 and -1 ) or 0;
end

--- Returns the total number of values passed
function math.count (...)
	return select("#", ...) or 0;
end

--- Returns the logarithm of the specified number to either ##e## or the specified base number.
function math.log (num, base)
	if base then
		return (log(num) / log(base or 10));
	else return log(num)
	end
end

--- Returns a list of all the unique numbers passed.
function math.unique(...)
    local tmp = {};
    local ret = {};
    for k, v in pairs({...}) do
        if ( tmp[v] == nil ) then
            tmp[v] = 1;
            table.insert(ret, v);
        end
    end
    return unpack(ret);
end

--- Returns the standard deviation for the population passed.
function math.stddev(...)
    local val = 0;
    local population = select("#", ...);
    for k, v in pairs({...}) do
        val = val + (tonumber(v) or 0);
    end
    local mean = val / population;
    local val = 0;
    for k, v in pairs({...}) do
        val = val + pow( ( (tonumber(v) or 0) - mean), 2);
    end
    val = val /(population-1);
    return sqrt(val);
end

--- Rounds the number to the nearest integer value.
function math.round(number, significance)
    return ( floor((number/(significance or 1))+0.5) * (significance or 1));
end

--- Rounds a number down (towards zero) to the nearest integer or multiple of significance.
function math.floor(number, significance)
    return ( floor((number/(significance or 1))) * (significance or 1));
end

--- Rounds a number up to the nearest integer value or the nearest multiple of significance.
function math.ceil(number, significance)
    return ( ceil((number/(significance or 1))) * (significance or 1));
end

--- Performs a matrix transformations on the coordinates passed
function math.matrix(x, y, t)
    if ( type(t) == "table" and #t > 0 ) then
		if ( #t == 6 ) then
			local nx = (x*t[1]) + (y*t[3]) + t[5];
			local ny = (x*t[2]) + (y*t[4]) + t[6];
			x,y = nx,ny;
		end
    end
    return x,y;
end

--- Performs a rotation of ##a## degrees of the coordinates passed by either 0,0 or cx,cy
function math.rotate (x,y,a,cx,cy)
	if not cx or cy then return math.matrix({cos(a),sin(a),-sin(a),cos(a),0,0},x,y)
	else
		x,y = math.matrix(x,y, {1, 0, 0, 1, -(cx or 0), -(cy or 0)});
        x,y = math.matrix(x,y, {cos(a),sin(a),-sin(a),cos(a),0,0});
		return math.matrix(x,y,{1, 0, 0, 1, (cx or 0), (cy or 0)});
	end
end

--- Skews the coordinates passed horizontally by ##a## degrees
function math.skewx (x,y,a) return math.matrix(x,y,{1, 0, tan(a), 1, 0, 0}) end

--- Skews the coordinates passed vertically by ##a## degrees
function math.skewy (x,y,a) return math.matrix(x,y,{1, tan(a), 0, 1, 0, 0}) end


